//
// Created by zx on 23-8-4.
//

#include "keyFramesManager.h"
#include <pcl/common/transforms.h>
keyFrame::keyFrame(PointCloudXYZI::Ptr cloud) {
  cloud_=cloud;
  cloud_world_.reset(new PointCloudXYZI());
  pose_changed_=true;
}
keyFrame::keyFrame(PointTypePose pose, PointCloudXYZI::Ptr cloud) {
  cloud_=cloud;
  cloud_world_.reset(new PointCloudXYZI());
  pose_=pose;
  pose_changed_=true;

}

keyFrame::keyFrame(const keyFrame& frame){
  cloud_=frame.cloud_;
  cloud_world_=frame.cloud_world_;
  pose_=frame.pose_;
  pose_changed_=frame.pose_changed_;
}
keyFrame& keyFrame::operator=(const keyFrame& frame){
  cloud_=frame.cloud_;
  cloud_world_=frame.cloud_world_;
  pose_=frame.pose_;
  pose_changed_=frame.pose_changed_;
  return *this;
}

keyFrame::~keyFrame() {}

void keyFrame::ResetPose(PointTypePose pose){
  lock_.lock();
  pose_=pose;
  pose_changed_=true;
  lock_.unlock();
}


PointCloudXYZI::Ptr keyFrame::world_cloud(Eigen::Isometry3d lidar_2_imu){
  if(pose_changed_==false){
    return cloud_world_;
  }
  lock_.lock();
  int cloudSize = cloud_->size();
  cloud_world_->resize(cloudSize);

  Eigen::Affine3f T_w_b_ = pcl::getTransformation(pose_.x, pose_.y, pose_.z, pose_.roll, pose_.pitch, pose_.yaw);

  Eigen::Isometry3d T_w_b ;          //   world2body
  T_w_b.matrix() = T_w_b_.matrix().cast<double>();

  Eigen::Isometry3d  T_w_lidar  =  T_w_b * lidar_2_imu  ;           //  T_w_lidar  转换矩阵

  Eigen::Isometry3d transCur = T_w_lidar;

  int numberOfCores=4;
#pragma omp parallel for num_threads(numberOfCores)
  for (int i = 0; i < cloudSize; ++i)
  {
    const auto &pointFrom = cloud_->points[i];
    cloud_world_->points[i].x = transCur(0, 0) * pointFrom.x + transCur(0, 1) * pointFrom.y + transCur(0, 2) * pointFrom.z + transCur(0, 3);
    cloud_world_->points[i].y = transCur(1, 0) * pointFrom.x + transCur(1, 1) * pointFrom.y + transCur(1, 2) * pointFrom.z + transCur(1, 3);
    cloud_world_->points[i].z = transCur(2, 0) * pointFrom.x + transCur(2, 1) * pointFrom.y + transCur(2, 2) * pointFrom.z + transCur(2, 3);
    cloud_world_->points[i].intensity = pointFrom.intensity;
  }
  pose_changed_=false;
  lock_.unlock();
  return cloud_world_;
}

//----------------------------------------------------------------------------------------------------

keyFramesManager::keyFramesManager(){}
keyFramesManager::~keyFramesManager(){}

pcl::PointCloud<PointTypePose>::Ptr keyFramesManager::GetPoses(){
  pcl::PointCloud<PointTypePose>::Ptr ret(new pcl::PointCloud<PointTypePose>);
  ret->resize(frames_.size());
  for(int i=0;i<frames_.size();++i){
    ret->points[i]=frames_[i].Pose();
  }
  return ret;
}